//Web server
#include "WiFi.h"
#include <HTTPClient.h>
#include <ESPAsyncWebServer.h>
#include <AsyncJson.h>

#include "Arduino.h"

//https://github.com/schreibfaul1/ESP32-audioI2S
#include "Audio.h" 
#include "SD.h"
#include "FS.h"
 
// Digital I/O used
#define SD_CS          5
#define SPI_MOSI      23    // SD Card
#define SPI_MISO      19
#define SPI_SCK       18

#define I2S_DOUT      22
#define I2S_BCLK      26
#define I2S_LRC       25

Audio audio;

HTTPClient http;

int replays = 0;

bool flagLocation = false;
bool flagGuide = false;

String startPoint;
String endPoint;


// Set web server port number to 32769
AsyncWebServer server(32769);

void notFound(AsyncWebServerRequest *request)
{
  request->send(404, "application/json", "{\"message\":\"Not found\"}");
}


void setup() {
  Serial.begin(115200);
  Serial.println("Init seccess!");
  
  //SSID and pwd
    /* 
  const char* ssid     = "RadiusAccessPoint";   //input your wifi name
  char* password = "3TristresTigres#";   //input your wifi passwords
  */  

  const char* ssid     = "Freebox-9B2771";   //input your wifi name
  char* password = "caveris*-juverunt-repressam7-circinent";   //input your wifi passwords


  WiFi.begin(ssid, password);
 
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi..");
  }
  
  Serial.println(WiFi.localIP());
  
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request) {
    request->send(200, "text/html", "<h1>ESP32 Web Server</h1><p>Sender-Receiver of the commands </p></body></html>");
  });

  server.on("/get-message", HTTP_GET, [](AsyncWebServerRequest *request) {
    StaticJsonDocument<100> data;
    if (request->hasParam("message"))
    {
      data["message"] = request->getParam("message")->value();
    }
    else
    {
      data["message"] = "No message parameter";
    }
    String response;
    serializeJson(data, response);
    request->send(200, "application/json", response);
  });
  
  AsyncCallbackJsonWebHandler *handler = new AsyncCallbackJsonWebHandler("/post-message", [](AsyncWebServerRequest *request, JsonVariant &json) {
    JsonObject jsonObj = json.as<JsonObject>();
    
    endPoint = jsonObj["endpoint"].as<String>();
    Serial.println(endPoint);

    String serverName = "http://tfgprojectjetson.ddns.net:32900/capture";

    http.begin(serverName.c_str());

    int httpResponseCode = http.GET();
  
    if (httpResponseCode>0) {
      Serial.print("HTTP Response code: ");
      Serial.println(httpResponseCode);
      String payload = http.getString();
      Serial.println(payload);
      
      DynamicJsonDocument doc(1024);

      deserializeJson(doc, payload);
      JsonObject obj = doc.as<JsonObject>();

      String audioName = obj["location"];
      startPoint = obj["startPoint"].as<String>();
      String url = "http://tfgprojectjetson.ddns.net:32900";
      String urlRequest = url + audioName;
      audio.connecttohost(urlRequest.c_str());
      flagLocation = true;
      flagGuide = true;
      request->send(200, "application/js  on", payload);
    }
    else {
      Serial.print("Error code: ");
      Serial.println(httpResponseCode);
    }
    // Free resources
    http.end();
  });
  
  server.addHandler(handler);
  server.onNotFound(notFound);
  server.begin();
  pinMode(SD_CS, OUTPUT);      digitalWrite(SD_CS, HIGH);
  SPI.begin(SPI_SCK, SPI_MISO, SPI_MOSI);
  Serial.begin(115200);
  SD.begin(SD_CS);

  audio.setPinout(I2S_BCLK, I2S_LRC, I2S_DOUT);
  audio.setVolume(21); // 0...21
}

void getInstructions(){
  String serverName = "http://tfgprojectjetson.ddns.net:32900/guide?startPoint=" + startPoint + "&endPoint=" + endPoint;
  Serial.println(serverName);

  http.begin(serverName.c_str());

  int httpResponseCode = http.GET();

  if (httpResponseCode>0) {
    Serial.print("HTTP Response code: ");
    Serial.println(httpResponseCode);
    String payload = http.getString();
    Serial.println(payload);
    
    DynamicJsonDocument doc(1024);

    deserializeJson(doc, payload);
    JsonObject obj = doc.as<JsonObject>();

    String audioName = obj["guide"];
    String url = "http://tfgprojectjetson.ddns.net:32900";
    String urlRequest = url + audioName;
    audio.connecttohost(urlRequest.c_str());
  }
  flagGuide = false;
}

void loop(){
  if(flagLocation == true) {
    audio.loop();
  }
}

// optional
void audio_info(const char *info){
    Serial.print("info        "); Serial.println(info);
    if (String(info).equals("slow stream, dropouts are possible")) {
      Serial.println("Detected ");
    }
}
void audio_id3data(const char *info){  //id3 metadata
    Serial.print("id3data     ");Serial.println(info);
}
void audio_eof_mp3(const char *info){  //end of file
    Serial.print("eof_mp3     ");Serial.println(info);
}
void audio_showstation(const char *info){
    Serial.print("station     ");Serial.println(info);
}
void audio_showstreamtitle(const char *info){
    Serial.print("streamtitle ");Serial.println(info);
}
void audio_bitrate(const char *info){
    Serial.print("bitrate     ");Serial.println(info);
}
void audio_commercial(const char *info){  //duration in sec
    Serial.print("commercial  ");Serial.println(info);
}
void audio_icyurl(const char *info){  //homepage
    Serial.print("icyurl      ");Serial.println(info);
}
void audio_lasthost(const char *info){  //stream URL played
  Serial.print("lasthost    ");Serial.println(info);
  replays++;
  if (replays == 2){
    audio.stopSong();
    replays = 0;
    Serial.println("Audio stopped");
    if (flagGuide == true){
      getInstructions();
    }
  }
}
void audio_eof_speech(const char *info){
    Serial.print("eof_speech  ");Serial.println(info);
}
