#https://gist.github.com/qpwo/cda55deee291de31b50d408c1a7c8515
import yaml
import json
from queue import PriorityQueue

class Graph:
    def __init__(self, num_of_vertices):
        self.v = num_of_vertices
        self.edges = [[-1 for i in range(num_of_vertices)] for j in range(num_of_vertices)]
        self.visited = []
        
    def add_edge(self, u, v, weight):
        self.edges[u][v] = weight
        self.edges[v][u] = weight

class Map:
    def __init__(self, fileName):
        self.visited = []
        with open(fileName, 'r') as file:
            self.data = json.loads(json.dumps(yaml.safe_load(file)))
        
        self.elements = {}
        self.n = 0
        self.edges = []
        self.dict = {}
        for node in self.data["Map"].keys():
            if node not in self.elements.keys():
                self.elements[node] = self.n
                self.n += 1
            self.dict[self.elements[node]] = set()
            for puerta in self.data["Map"][node]:
                print(node + " -> " + puerta)
                if puerta not in self.elements.keys():
                    self.elements[puerta] = self.n
                    self.n += 1
                element = (self.elements[node], self.elements[puerta])
                self.edges.append(element)
                self.dict[self.elements[node]].add((self.elements[puerta], 1))
            
        #print(self.edges)
        # asigna memoria para la lista de adyacencia
        self.adjList = [[] for _ in range(self.n)]
 
        self.g = Graph(self.n)
        # agrega bordes al graph dirigido
        for (src, dest) in self.edges:
            # asignar nodo en la lista de adyacencia de src a dest
            self.adjList[src].append(dest)
            self.g.add_edge(src, dest, 1)
            
        print(self.adjList)
 
    # Función para imprimir la representación de lista de adyacencia de un graph
    def printGraph(self, graph):
        for src in range(len(graph.adjList)):
            # imprime el vértice actual y todos sus vértices vecinos
            for dest in graph.adjList[src]:
                print(f'({src} —> {dest}) ', end='')
            print()
    

    def dijkstra(self, G, start, goal):
        """ Uniform-cost search / dijkstra """
        visited = set()
        cost = {start: 0}
        parent = {start: None}
        todo = PriorityQueue()
    
        todo.put((0, start))
        while todo:
            while not todo.empty():
                _, vertex = todo.get() # finds lowest cost vertex
                # loop until we get a fresh vertex
                if vertex not in visited: break
            else: # if todo ran out
                break # quit main loop
            visited.add(vertex)
            if vertex == goal:
                break
            for neighbor, distance in G[vertex]:
                if neighbor in visited: continue # skip these to save time
                old_cost = cost.get(neighbor, float('inf')) # default to infinity
                new_cost = cost[vertex] + distance
                if new_cost < old_cost:
                    todo.put((new_cost, neighbor))
                    cost[neighbor] = new_cost
                    parent[neighbor] = vertex

        return parent

    def make_path(self, parent, goal):
        if goal not in parent:
            return None
        v = goal
        path = []
        while v is not None: # root has null parent
            path.append(v)
            v = parent[v]
        return path[::-1]
 