# webserver.py
from flask import Flask, request, Response
import cv2
import logging

import torch
from torch.autograd import Variable as V
import torchvision.models as models
from torchvision import transforms as trn
from torch.nn import functional as F
import os
import numpy as np
from PIL import Image

from path import indoormap
features_blobs = []

# hacky way to deal with the Pytorch 1.0 update
def recursion_change_bn(module):
    if isinstance(module, torch.nn.BatchNorm2d):
        module.track_running_stats = 1
    else:
        for i, (name, module1) in enumerate(module._modules.items()):
            module1 = recursion_change_bn(module1)
    return module

def load_labels():
    # prepare all the labels
    # scene category relevant
    file_name_category = 'categories_places365.txt'
    if not os.access(file_name_category, os.W_OK):
        synset_url = 'https://raw.githubusercontent.com/csailvision/places365/master/categories_places365.txt'
        os.system('wget ' + synset_url)
    classes = list()
    with open(file_name_category) as class_file:
        for line in class_file:
            classes.append(line.strip().split(' ')[0][3:])
    classes = tuple(classes)

    # indoor and outdoor relevant
    file_name_IO = 'IO_places365.txt'
    if not os.access(file_name_IO, os.W_OK):
        synset_url = 'https://raw.githubusercontent.com/csailvision/places365/master/IO_places365.txt'
        os.system('wget ' + synset_url)
    with open(file_name_IO) as f:
        lines = f.readlines()
        labels_IO = []
        for line in lines:
            items = line.rstrip().split()
            labels_IO.append(int(items[-1]) -1) # 0 is indoor, 1 is outdoor
    labels_IO = np.array(labels_IO)

    # scene attribute relevant
    file_name_attribute = 'labels_sunattribute.txt'
    if not os.access(file_name_attribute, os.W_OK):
        synset_url = 'https://raw.githubusercontent.com/csailvision/places365/master/labels_sunattribute.txt'
        os.system('wget ' + synset_url)
    with open(file_name_attribute) as f:
        lines = f.readlines()
        labels_attribute = [item.rstrip() for item in lines]
    file_name_W = 'W_sceneattribute_wideresnet18.npy'
    if not os.access(file_name_W, os.W_OK):
        synset_url = 'http://places2.csail.mit.edu/models_places365/W_sceneattribute_wideresnet18.npy'
        os.system('wget ' + synset_url)
    W_attribute = np.load(file_name_W)

    return classes, labels_IO, labels_attribute, W_attribute

def hook_feature(module, input, output):
    features_blobs.append(np.squeeze(output.data.cpu().numpy()))

def returnCAM(feature_conv, weight_softmax, class_idx):
    # generate the class activation maps upsample to 256x256
    size_upsample = (256, 256)
    nc, h, w = feature_conv.shape
    output_cam = []
    for idx in class_idx:
        cam = weight_softmax[class_idx].dot(feature_conv.reshape((nc, h*w)))
        cam = cam.reshape(h, w)
        cam = cam - np.min(cam)
        cam_img = cam / np.max(cam)
        cam_img = np.uint8(255 * cam_img)
        output_cam.append(cv2.resize(cam_img, size_upsample))
    return output_cam

def returnTF():
# load the image transformer
    tf = trn.Compose([
        trn.Resize((224,224)),
        trn.ToTensor(),
        trn.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    ])
    return tf


def load_model():
    # this model has a last conv feature map as 14x14

    #model_file = 'wideresnet18_places365.pth.tar'
    model_file = 'resnet18_best.pth.tar'

    if not os.access(model_file, os.W_OK):
        os.system('wget http://places2.csail.mit.edu/models_places365/' + model_file)
        os.system('wget https://raw.githubusercontent.com/csailvision/places365/master/wideresnet.py')

    import wideresnet
    #model = wideresnet.resnet18(num_classes=365)
    model = wideresnet.resnet18(num_classes=6)
    checkpoint = torch.load(model_file, map_location=lambda storage, loc: storage)
    state_dict = {str.replace(k,'module.',''): v for k,v in checkpoint['state_dict'].items()}
    model.load_state_dict(state_dict)
    
    # hacky way to deal with the upgraded batchnorm2D and avgpool layers...
    for i, (name, module) in enumerate(model._modules.items()):
        module = recursion_change_bn(model)
    model.avgpool = torch.nn.AvgPool2d(kernel_size=14, stride=1, padding=0)
    
    model.eval()



    # the following is deprecated, everything is migrated to python36

    ## if you encounter the UnicodeDecodeError when use python3 to load the model, add the following line will fix it. Thanks to @soravux
    #from functools import partial
    #import pickle
    #pickle.load = partial(pickle.load, encoding="latin1")
    #pickle.Unpickler = partial(pickle.Unpickler, encoding="latin1")
    #model = torch.load(model_file, map_location=lambda storage, loc: storage, pickle_module=pickle)

    model.eval()
    # hook the feature extractor
    features_names = ['layer4','avgpool'] # this is the last conv layer of the resnet
    for name in features_names:
        model._modules.get(name).register_forward_hook(hook_feature)
    return model

app = Flask(__name__)

@app.get("/capture")
def get_capture():
    cam = cv2.VideoCapture("http://tfgprojectjetson.ddns.net:32770/stream")

    img_counter = 0

    ret, frame = cam.read()
    img_name = ""
    if not ret:
        img_name = "failed to grab frame"
    else:
        img_name = "opencv_frame_{}.png".format(img_counter)
        cv2.imwrite(img_name, frame)
        app.logger.warning("{} written!".format(img_name))
        img_counter += 1

    cam.release()

    probs, classes, idx = prediction(frame, img_name)

    os.system(f'espeak -v en-gb "With a probability of {round(probs[0]*100)} you are located in the {classes[idx[0]]}" -w audios/command.wav')
    os.system('lame audios/command.wav audios/command.mp3')

    return {"startPoint": classes[idx[0]], "location":f"/audios?audio=command.mp3"}, 200

@app.get("/audios")
def getAudio():
    audio = request.args.get('audio')   
    def generate():
        with open(f"audios/{audio}", "rb") as fwav:
            data = fwav.read(1024)
            while data:
                yield data
                data = fwav.read(1024)
    return Response(generate(), mimetype="audio/mp3")

@app.get("/audioGuide")
def getGuide():
    audio = request.args.get('audio')   
    def generate():
        with open(f"audios/{audio}", "rb") as fwav:
            data = fwav.read(1024)
            while data:
                yield data
                data = fwav.read(1024)
    return Response(generate(), mimetype="audio/mp3")


@app.get("/guide")
def getPath():
    startPoint = request.args.get('startPoint') 
    endPoint = request.args.get('endPoint')     

    graph = indoormap.Map("path/map.yml")
    graph.printGraph(graph)

    key_list = list(graph.elements.keys())
    val_list = list(graph.elements.values())

    app.logger.warning(startPoint)
    app.logger.warning(endPoint)


    parent = graph.dijkstra(graph.dict, key_list.index(startPoint), key_list.index(endPoint))
    finalPath = graph.make_path(parent, key_list.index(endPoint)) 
    
    # print key with val 100
    print("The shortest path from: " + startPoint + " To: " + endPoint + " Is: ")
    camino = []
    comando = f"In order to go from {startPoint} to {endPoint}, you need to go to the following places: "
    for i in finalPath:
        print(key_list[val_list.index(i)])
        camino.append(key_list[val_list.index(i)])
        comando = comando + key_list[val_list.index(i)] + ". "
    
    os.system(f'espeak -v en-gb "{comando}" -w audios/guide.wav')
    os.system('lame audios/guide.wav audios/guide.mp3')

    return {"guide":"/audioGuide?audio=guide.mp3"}, 200

def prediction(frame, img_name):
    # load the labels
    classes, labels_IO, labels_attribute, W_attribute = load_labels()

    # load the model
    app.logger.warning("Loading model")
    model = load_model()
    # load the transformer
    app.logger.warning("Loading transformer")
    tf = returnTF() # image transformer

    # get the softmax weight
    app.logger.warning("Getting weights")
    params = list(model.parameters())
    weight_softmax = params[-2].data.numpy()
    weight_softmax[weight_softmax<0] = 0

    # load the test image
    app.logger.warning("Loading image")
    
    #Converting to PIL imaged (added by me)
    img = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    img = Image.fromarray(img)

    input_img = V(tf(img).unsqueeze(0))

    # forward pass
    logit = model.forward(input_img)
    h_x = F.softmax(logit, 1).data.squeeze()
    probs, idx = h_x.sort(0, True)
    probs = probs.numpy()
    idx = idx.numpy()

    app.logger.warning('RESULT ON ' + img_name)

    # output the IO prediction
    io_image = np.mean(labels_IO[idx[:10]]) # vote for the indoor or outdoor
    if io_image < 0.5:
        app.logger.warning('--TYPE OF ENVIRONMENT: indoor')
    else:
        app.logger.warning('--TYPE OF ENVIRONMENT: outdoor')

    # output the prediction of scene category
    app.logger.warning('--SCENE CATEGORIES:')
    for i in range(0, 5):
        app.logger.warning('{:.3f} -> {}'.format(probs[i], classes[idx[i]]))

    # output the scene attributes
    responses_attribute = W_attribute.dot(features_blobs[1])
    idx_a = np.argsort(responses_attribute)
    app.logger.warning('--SCENE ATTRIBUTES:')
    app.logger.warning(', '.join([labels_attribute[idx_a[i]] for i in range(-1,-10,-1)]))

    # generate class activation mapping
    app.logger.warning('Class activation map is saved as cam.jpg')
    CAMs = returnCAM(features_blobs[0], weight_softmax, [idx[0]])

    # render the CAM and output
    img = cv2.imread(img_name)
    height, width, _ = img.shape
    heatmap = cv2.applyColorMap(cv2.resize(CAMs[0],(width, height)), cv2.COLORMAP_JET)
    result = heatmap * 0.4 + img * 0.5
    cv2.imwrite('cam.jpg', result)

    return [probs, classes, idx]

